@@ -8,24 +8,25 @@
 # TODO: - wrong versions in versionadded/changed
 #       - wrong markup after versionchanged directive
 
-"""Sphinx rst linter.
-"""
+"""Sphinx rst linter."""
 
-__version__ = "0.6.1"
+__version__ = "0.6.3"
 
 import argparse
 import multiprocessing
 import os
-import re
 import sys
 from collections import Counter
 from functools import reduce
 from itertools import chain, starmap
 from os.path import exists, isfile, join, splitext
 
+import regex as re
+
+
 # The following chars groups are from docutils:
-closing_delimiters = "\\\\.,;!?"
-delimiters = (
+CLOSING_DELIMITERS = "\\\\.,;!?"
+DELIMITERS = (
     "\\-/:\u058a\xa1\xb7\xbf\u037e\u0387\u055a-\u055f\u0589"
     "\u05be\u05c0\u05c3\u05c6\u05f3\u05f4\u0609\u060a\u060c"
     "\u060d\u061b\u061e\u061f\u066a-\u066d\u06d4\u0700-\u070d"
@@ -47,7 +48,8 @@ delimiters = (
     "\ufe6b\uff01-\uff03\uff05-\uff07\uff0a\uff0c-\uff0f\uff1a"
     "\uff1b\uff1f\uff20\uff3c\uff61\uff64\uff65"
 )
-closers = (
+
+CLOSERS = (
     "\"')>\\]}\u0f3b\u0f3d\u169c\u2046\u207e\u208e\u232a\u2769"
     "\u276b\u276d\u276f\u2771\u2773\u2775\u27c6\u27e7\u27e9\u27eb"
     "\u27ed\u27ef\u2984\u2986\u2988\u298a\u298c\u298e\u2990\u2992"
@@ -59,7 +61,8 @@ closers = (
     "\u201b\u201f\xab\u2018\u201c\u2039\u2e02\u2e04\u2e09\u2e0c"
     "\u2e1c\u2e20\u201a\u201e"
 )
-openers = (
+
+OPENERS = (
     "\"'(<\\[{\u0f3a\u0f3c\u169b\u2045\u207d\u208d\u2329\u2768"
     "\u276a\u276c\u276e\u2770\u2772\u2774\u27c5\u27e6\u27e8\u27ea"
     "\u27ec\u27ee\u2983\u2985\u2987\u2989\u298b\u298d\u298f\u2991"
@@ -72,9 +75,8 @@ openers = (
     "\u2e1d\u2e21\u201b\u201f"
 )
 
-
 # fmt: off
-directives = [
+DIRECTIVES = [
     # standard docutils ones
     'admonition', 'attention', 'caution', 'class', 'compound', 'container',
     'contents', 'csv-table', 'danger', 'date', 'default-role', 'epigraph',
@@ -101,10 +103,11 @@ directives = [
 # fmt: on
 
 
-all_directives = "(" + "|".join(directives) + ")"
-before_role = r"(^|(?<=[\s(/'{\[*-]))"
-simplename = r"(?:(?!_)\w)+(?:[-._+:](?:(?!_)\w)+)*"
-role_head = rf"({before_role}:{simplename}:)"  # A role, with a clean start
+ALL_DIRECTIVES = "(" + "|".join(DIRECTIVES) + ")"
+BEFORE_ROLE = r"(^|(?<=[\s(/'{\[*-]))"
+SIMPLENAME = r"(?:(?!_)\w)+(?:[-._+:](?:(?!_)\w)+)*"
+ROLE_TAG = rf":{SIMPLENAME}:"
+ROLE_HEAD = rf"({BEFORE_ROLE}:{SIMPLENAME}:)"  # A role, with a clean start
 
 # Find comments that look like a directive, like:
 # .. versionchanged 3.6
@@ -112,22 +115,22 @@ role_head = rf"({before_role}:{simplename}:)"  # A role, with a clean start
 # .. versionchanged: 3.6
 # as it should be:
 # .. versionchanged:: 3.6
-seems_directive_re = re.compile(rf"^\s*(?<!\.)\.\. {all_directives}([^a-z:]|:(?!:))")
+seems_directive_re = re.compile(rf"^\s*(?<!\.)\.\. {ALL_DIRECTIVES}([^a-z:]|:(?!:))")
 
 # Find directive prefixed with three dots instead of two, like:
 # ... versionchanged:: 3.6
 # instead of:
 # .. versionchanged:: 3.6
-three_dot_directive_re = re.compile(rf"\.\.\. {all_directives}::")
+three_dot_directive_re = re.compile(rf"\.\.\. {ALL_DIRECTIVES}::")
 
 # Find role used with double backticks instead of simple backticks like:
 # :const:``None``
 # instead of:
 # :const:`None`
-double_backtick_role = re.compile(rf"(?<!``){role_head}``")
+double_backtick_role = re.compile(rf"(?<!``){ROLE_HEAD}``")
 
-start_string_prefix = f"(^|(?<=\\s|[{openers}{delimiters}|]))"
-end_string_suffix = f"($|(?=\\s|[\x00{closing_delimiters}{delimiters}{closers}|]))"
+start_string_prefix = f"(^|(?<=\\s|[{OPENERS}{DELIMITERS}|]))"
+end_string_suffix = f"($|(?=\\s|[\x00{CLOSING_DELIMITERS}{DELIMITERS}{CLOSERS}|]))"
 
 # Find role glued with another word like:
 #     the:c:func:`PyThreadState_LeaveTracing` function.
@@ -138,18 +141,14 @@ end_string_suffix = f"($|(?=\\s|[\x00{closing_delimiters}{delimiters}{closers}|]
 #     issue:`123`
 # instead of:
 #     :issue:`123`
-role_glued_with_word = re.compile(rf"(^|\s)(?!:){simplename}:`(?!`)")
+role_glued_with_word = re.compile(rf"(^|\s)(?!:){SIMPLENAME}:`(?!`)")
 
 
-role_with_no_backticks = re.compile(rf"(^|\s):{simplename}:(?![`:])[^\s`]+(\s|$)")
+role_with_no_backticks = re.compile(rf"(^|\s):{SIMPLENAME}:(?![`:])[^\s`]+(\s|$)")
 
 # Find role missing middle colon, like:
 #    The :issue`123` is ...
-role_missing_right_colon = re.compile(rf"(^|\s):{simplename}`(?!`)")
-
-# TODO: cover more cases
-# https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#toc-entry-44
-default_role_re = re.compile(r"(^| )`\w([^`]*?\w)?`($| )")
+role_missing_right_colon = re.compile(rf"(^|\s):{SIMPLENAME}`(?!`)")
 
 seems_hyperlink_re = re.compile(r"`[^`]+?(\s?)<https?://[^`]+>`(_?)")
 
@@ -190,11 +189,7 @@ def check_python_syntax(file, lines, options=None):
         yield err.lineno, f"not compilable: {err}"
 
 
-def is_in_a_table(error, line):
-    return "|" in line[: error.start()] and "|" in line[error.end() :]
-
-
-role_missing_closing_backtick = re.compile(rf"({role_head}`[^`]+?)[^`]*$")
+role_missing_closing_backtick = re.compile(rf"({ROLE_HEAD}`[^`]+?)[^`]*$")
 
 
 @checker(".rst")
@@ -223,10 +218,10 @@ def check_missing_space_after_literal(file, lines, options=None):
     for paragraph_lno, paragraph in paragraphs(lines):
         if paragraph.count("|") > 4:
             return  # we don't handle tables yet.
-        paragraph_without_roles = re.sub(normal_role, "", paragraph).replace("````", "")
-        for role in re.finditer(
-            "``.+?``(?!`).", paragraph_without_roles, flags=re.DOTALL
-        ):
+        paragraph = escape2null(paragraph)
+        paragraph = inline_literal_re.sub("", paragraph)
+        paragraph = normal_role_re.sub("", paragraph)
+        for role in re.finditer("``.+?``(?!`).", paragraph, flags=re.DOTALL):
             if not re.match(end_string_suffix, role.group(0)[-1]):
                 error_offset = paragraph[: role.start()].count("\n")
                 yield (
@@ -236,6 +231,60 @@ def check_missing_space_after_literal(file, lines, options=None):
                 )
 
 
+@checker(".rst")
+def check_unbalanced_inline_literals_delimiters(file, lines, options=None):
+    r"""Search for unbalanced inline literals delimiters.
+
+    Bad:  ``hello`` world``
+    Good: ``hello`` world
+    """
+    for paragraph_lno, paragraph in paragraphs(lines):
+        if paragraph.count("|") > 4:
+            return  # we don't handle tables yet.
+        paragraph = escape2null(paragraph)
+        paragraph = inline_literal_re.sub("", paragraph)
+        paragraph = normal_role_re.sub("", paragraph)
+        for lone_double_backtick in re.finditer("(?<!`)``(?!`)", paragraph):
+            error_offset = paragraph[: lone_double_backtick.start()].count("\n")
+            yield (
+                paragraph_lno + error_offset,
+                "found an unbalanced inline literal markup.",
+            )
+
+
+def escape2null(text):
+    r"""Return a string with escape-backslashes converted to nulls.
+
+    It ease telling appart escaping-backslashes and normal backslashes
+    in regex.
+
+    For example : \\\\\\` is hard to match, even with the eyes, it's
+    hard to know which backslash escapes which backslash, and it's
+    very hard to know if the backtick is escaped.
+
+    By replacing the escaping backslashes with another character they
+    become easy to spot:
+
+    0\0\0\`
+
+    (This example uses zeros for readability but the function actually
+    uses null bytes, \x00.)
+
+    So we easily see that the backtick is **not** escaped: it's
+    preceded by a backslash, not an escaping backslash.
+    """
+    parts = []
+    start = 0
+    while True:
+        found = text.find("\\", start)
+        if found == -1:
+            parts.append(text[start:])
+            return "".join(parts)
+        parts.append(text[start:found])
+        parts.append("\x00" + text[found + 1 : found + 2])
+        start = found + 2  # skip character after escape
+
+
 def paragraphs(lines):
     """Yield (paragraph_line_no, paragraph_text) pairs describing
     paragraphs of the given lines.
@@ -255,9 +304,85 @@ def paragraphs(lines):
         yield paragraph_lno, "".join(paragraph)
 
 
-role_body = rf"([^`]|\s`+|\\`|:{simplename}:`([^`]|\s`+|\\`)+`)+"
-normal_role = f":{simplename}:`{role_body}`"
-backtick_in_front_of_role = re.compile(rf"(^|\s)`:{simplename}:`{role_body}`")
+QUOTE_PAIRS = [
+    "»»",  # Swedish
+    "‘‚",  # Albanian/Greek/Turkish
+    "’’",  # Swedish
+    "‚‘",  # German
+    "‚’",  # Polish
+    "“„",  # Albanian/Greek/Turkish
+    "„“",  # German
+    "„”",  # Polish
+    "””",  # Swedish
+    "››",  # Swedish
+    "''",  # ASCII
+    '""',  # ASCII
+    "<>",  # ASCII
+    "()",  # ASCII
+    "[]",  # ASCII
+    "{}",  # ASCII
+]
+
+QUOTE_PAIRS_NEGATIVE_LOOKBEHIND = (
+    "(?<!"
+    + "|".join(f"{re.escape(pair[0])}`{re.escape(pair[1])}" for pair in QUOTE_PAIRS)
+    + "|"
+    + "|".join(
+        f"{opener}`{closer}"
+        for opener, closer in zip(map(re.escape, OPENERS), map(re.escape, CLOSERS))
+    )
+    + ")"
+)
+
+
+def inline_markup_gen(start_string, end_string):
+    """Generate a regex matching an inline markup.
+
+    inline_markup_gen('**', '**') geneates a regex matching strong
+    emphasis inline markup.
+    """
+    return re.compile(
+        r"""
+    (?<!\x00) # Both inline markup start-string and end-string must not be preceded by an unescaped backslash
+    (?<=      # Inline markup start-strings must:
+        ^|           # start a text block
+        \s|          # or be immediately preceded by whitespace,
+        [-:/'"<([{]| # one of the ASCII characters
+        [\p{Ps}\p{Pi}\p{Pf}\p{Pd}\p{Po}]  # or a similar non-ASCII punctuation character.
+    )
+    (?P<inline_markup>"""
+        + start_string
+        + r"""        # Inline markup start
+        \S     # Inline markup start-strings must be immediately followed by non-whitespace.
+               # The inline markup end-string must be separated by at least one character from the start-string.
+        """
+        + QUOTE_PAIRS_NEGATIVE_LOOKBEHIND
+        + r"""
+        .*?
+    """
+        + end_string
+        + r""")        # Inline markup end
+    (?=       # Inline markup end-strings must
+        $|    # end a text block or
+        \s|   # be immediately followed by whitespace,
+        \x00|
+        [-.,:;!?/'")\]}>]|  # one of the ASCII characters
+        [\p{Pe}\p{Pi}\p{Pf}\p{Pd}\p{Po}]  # or a similar non-ASCII punctuation character.
+    )
+    """,
+        flags=re.VERBOSE | re.DOTALL,
+    )
+
+
+# https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#inline-markup-recognition-rules
+interpreted_text_re = inline_markup_gen("`", "`")
+inline_literal_re = inline_markup_gen("``", "``")
+normal_role_re = re.compile(
+    f":{SIMPLENAME}:{interpreted_text_re.pattern}", flags=re.VERBOSE | re.DOTALL
+)
+backtick_in_front_of_role = re.compile(
+    rf"(^|\s)`:{SIMPLENAME}:{interpreted_text_re.pattern}"
+)
 
 
 @checker(".rst", enabled=False)
@@ -268,7 +393,18 @@ def check_default_role(file, lines, options=None):
     Good: ``print``
     """
     for lno, line in enumerate(lines, start=1):
-        if default_role_re.search(line):
+        line = escape2null(line)
+        line = inline_literal_re.sub("", line)
+        match = interpreted_text_re.search(line)
+        if match:
+            before_match = line[: match.start()]
+            after_match = line[match.end() :]
+            if re.search(ROLE_TAG + "$", before_match):
+                # It's not a default role: it starts with a tag.
+                continue
+            if re.search("^" + ROLE_TAG, after_match):
+                # It's not a default role: it ends with a tag.
+                continue
             yield lno, "default role used (hint: for inline literals, use double backticks)"
 
 
@@ -307,9 +443,12 @@ def check_missing_space_after_role(file, lines, options=None):
     #    The :literal:`:exc:`Exceptions``
     # While this is not:
     #    The :literal:`:exc:`Exceptions``s
+    role_body = rf"([^`]|\s`+|\\`|:{SIMPLENAME}:`([^`]|\s`+|\\`)+`)+"
+    suspicious_role = re.compile(f":{SIMPLENAME}:`{role_body}`s")
     for lno, line in enumerate(lines, start=1):
-        line = re.sub("``.*?``", "", line).replace("````", "")
-        role = re.search(rf"{normal_role}s", line)
+        line = inline_literal_re.sub("", line)
+        line = normal_role_re.sub("", line)
+        role = suspicious_role.search(line)
         if role:
             yield lno, f"role missing (escaped) space after role: {role.group(0)!r}"
 
@@ -474,6 +613,7 @@ def check_leaked_markup(file, lines, options=None):
 
 
 def is_multiline_non_rst_block(line):
+    """Returns True if the next lines are an indented literal block."""
     if line.endswith("..\n"):
         return True
     if line.endswith("::\n"):
@@ -519,7 +659,8 @@ def hide_non_rst_blocks(lines, hidden_block_cb=None):
 
 
 def type_of_explicit_markup(line):
-    if re.match(rf"\.\. {all_directives}::", line):
+    """Tell apart various explicit markup blocks."""
+    if re.match(rf"\.\. {ALL_DIRECTIVES}::", line):
         return "directive"
     if re.match(r"\.\. \[[0-9]+\] ", line):
         return "footnote"
@@ -579,6 +720,7 @@ def check_bad_dedent(file, lines, options=None):
 
 
 def parse_args(argv=None):
+    """Parse command line argument."""
     if argv is None:
         argv = sys.argv
     parser = argparse.ArgumentParser(description=__doc__)
@@ -661,6 +803,8 @@ def walk(path, ignore_list):
     It also allows giving a file, thus yielding just that file.
     """
     if isfile(path):
+        if path in ignore_list:
+            return
         yield path if path[:2] != "./" else path[2:]
         return
     for root, dirs, files in os.walk(path):
